@@ -9,32 +9,21 @@
 # obtain one at https://mozilla.org/MPL/2.0/.
 
 import datetime as dt
-import os.path
+import operator as op
+import zoneinfo
 from calendar import monthrange
-from functools import lru_cache
+from functools import cache, partial
 from importlib import resources
+from pathlib import Path
 from typing import Optional
 
 from hypothesis.errors import InvalidArgument
-from hypothesis.internal.conjecture import utils
 from hypothesis.internal.validation import check_type, check_valid_interval
 from hypothesis.strategies._internal.core import sampled_from
-from hypothesis.strategies._internal.misc import just, none
+from hypothesis.strategies._internal.misc import just, none, nothing
 from hypothesis.strategies._internal.strategies import SearchStrategy
 from hypothesis.strategies._internal.utils import defines_strategy
 
-# The zoneinfo module, required for the timezones() and timezone_keys()
-# strategies, is new in Python 3.9 and the backport might be missing.
-try:
-    import zoneinfo
-except ImportError:
-    try:
-        from backports import zoneinfo  # type: ignore
-    except ImportError:
-        # We raise an error recommending `pip install hypothesis[zoneinfo]`
-        # when timezones() or timezone_keys() strategies are actually used.
-        zoneinfo = None  # type: ignore
-
 DATENAMES = ("year", "month", "day")
 TIMENAMES = ("hour", "minute", "second", "microsecond")
 
@@ -113,9 +102,9 @@ def draw_capped_multipart(
         if name == "day" and not cap_high:
             _, high = monthrange(**result)
         if name == "year":
-            val = utils.integer_range(data, low, high, 2000)
+            val = data.draw_integer(low, high, shrink_towards=2000)
         else:
-            val = utils.integer_range(data, low, high)
+            val = data.draw_integer(low, high)
         result[name] = val
         cap_low = cap_low and val == low
         cap_high = cap_high and val == high
@@ -125,12 +114,13 @@ def draw_capped_multipart(
         # the logic above, and be very sensitive to the specific timezone
         # (at the cost of efficient shrinking and mutation), so at least for
         # now we stick with the status quo and generate it independently.
-        result["fold"] = utils.integer_range(data, 0, 1)
+        result["fold"] = data.draw_integer(0, 1)
     return result
 
 
 class DatetimeStrategy(SearchStrategy):
     def __init__(self, min_value, max_value, timezones_strat, allow_imaginary):
+        super().__init__()
         assert isinstance(min_value, dt.datetime)
         assert isinstance(max_value, dt.datetime)
         assert min_value.tzinfo is None
@@ -156,7 +146,7 @@ class DatetimeStrategy(SearchStrategy):
 
         # If we happened to end up with a disallowed imaginary time, reject it.
         if (not self.allow_imaginary) and datetime_does_not_exist(result):
-            data.mark_invalid()
+            data.mark_invalid(f"{result} does not exist (usually a DST transition)")
         return result
 
     def draw_naive_datetime_and_combine(self, data, tz):
@@ -164,9 +154,10 @@ class DatetimeStrategy(SearchStrategy):
         try:
             return replace_tzinfo(dt.datetime(**result), timezone=tz)
         except (ValueError, OverflowError):
-            msg = "Failed to draw a datetime between %r and %r with timezone from %r."
-            data.note_event(msg % (self.min_value, self.max_value, self.tz_strat))
-            data.mark_invalid()
+            data.mark_invalid(
+                f"Failed to draw a datetime between {self.min_value!r} and "
+                f"{self.max_value!r} with timezone from {self.tz_strat!r}."
+            )
 
 
 @defines_strategy(force_reusable_values=True)
@@ -189,8 +180,7 @@ def datetimes(
     You can construct your own, though we recommend using one of these built-in
     strategies:
 
-    * with Python 3.9 or newer or :pypi:`backports.zoneinfo`:
-      :func:`hypothesis.strategies.timezones`;
+    * with the standard library: :func:`hypothesis.strategies.timezones`;
     * with :pypi:`dateutil <python-dateutil>`:
       :func:`hypothesis.extra.dateutil.timezones`; or
     * with :pypi:`pytz`: :func:`hypothesis.extra.pytz.timezones`.
@@ -216,13 +206,13 @@ def datetimes(
     check_type(dt.datetime, min_value, "min_value")
     check_type(dt.datetime, max_value, "max_value")
     if min_value.tzinfo is not None:
-        raise InvalidArgument(f"min_value={min_value!r} must not have tzinfo")
+        raise InvalidArgument(f"{min_value=} must not have tzinfo")
     if max_value.tzinfo is not None:
-        raise InvalidArgument(f"max_value={max_value!r} must not have tzinfo")
+        raise InvalidArgument(f"{max_value=} must not have tzinfo")
     check_valid_interval(min_value, max_value, "min_value", "max_value")
     if not isinstance(timezones, SearchStrategy):
         raise InvalidArgument(
-            f"timezones={timezones!r} must be a SearchStrategy that can "
+            f"{timezones=} must be a SearchStrategy that can "
             "provide tzinfo for datetimes (either None or dt.tzinfo objects)"
         )
     return DatetimeStrategy(min_value, max_value, timezones, allow_imaginary)
@@ -230,6 +220,7 @@ def datetimes(
 
 class TimeStrategy(SearchStrategy):
     def __init__(self, min_value, max_value, timezones_strat):
+        super().__init__()
         self.min_value = min_value
         self.max_value = max_value
         self.tz_strat = timezones_strat
@@ -259,15 +250,16 @@ def times(
     check_type(dt.time, min_value, "min_value")
     check_type(dt.time, max_value, "max_value")
     if min_value.tzinfo is not None:
-        raise InvalidArgument(f"min_value={min_value!r} must not have tzinfo")
+        raise InvalidArgument(f"{min_value=} must not have tzinfo")
     if max_value.tzinfo is not None:
-        raise InvalidArgument(f"max_value={max_value!r} must not have tzinfo")
+        raise InvalidArgument(f"{max_value=} must not have tzinfo")
     check_valid_interval(min_value, max_value, "min_value", "max_value")
     return TimeStrategy(min_value, max_value, timezones)
 
 
 class DateStrategy(SearchStrategy):
     def __init__(self, min_value, max_value):
+        super().__init__()
         assert isinstance(min_value, dt.date)
         assert isinstance(max_value, dt.date)
         assert min_value < max_value
@@ -279,6 +271,37 @@ class DateStrategy(SearchStrategy):
             **draw_capped_multipart(data, self.min_value, self.max_value, DATENAMES)
         )
 
+    def filter(self, condition):
+        if (
+            isinstance(condition, partial)
+            and len(args := condition.args) == 1
+            and not condition.keywords
+            and isinstance(arg := args[0], dt.date)
+            and condition.func in (op.lt, op.le, op.eq, op.ge, op.gt)
+        ):
+            try:
+                arg += dt.timedelta(days={op.lt: 1, op.gt: -1}.get(condition.func, 0))
+            except OverflowError:  # gt date.max, or lt date.min
+                return nothing()
+            lo, hi = {
+                # We're talking about op(arg, x) - the reverse of our usual intuition!
+                op.lt: (arg, self.max_value),  # lambda x: arg < x
+                op.le: (arg, self.max_value),  # lambda x: arg <= x
+                op.eq: (arg, arg),  #            lambda x: arg == x
+                op.ge: (self.min_value, arg),  # lambda x: arg >= x
+                op.gt: (self.min_value, arg),  # lambda x: arg > x
+            }[condition.func]
+            lo = max(lo, self.min_value)
+            hi = min(hi, self.max_value)
+            print(lo, hi)
+            if hi < lo:
+                return nothing()
+            if lo <= self.min_value and self.max_value <= hi:
+                return self
+            return dates(lo, hi)
+
+        return super().filter(condition)
+
 
 @defines_strategy(force_reusable_values=True)
 def dates(
@@ -300,6 +323,7 @@ def dates(
 
 class TimedeltaStrategy(SearchStrategy):
     def __init__(self, min_value, max_value):
+        super().__init__()
         assert isinstance(min_value, dt.timedelta)
         assert isinstance(max_value, dt.timedelta)
         assert min_value < max_value
@@ -313,7 +337,7 @@ class TimedeltaStrategy(SearchStrategy):
         for name in ("days", "seconds", "microseconds"):
             low = getattr(self.min_value if low_bound else dt.timedelta.min, name)
             high = getattr(self.max_value if high_bound else dt.timedelta.max, name)
-            val = utils.integer_range(data, low, high, 0)
+            val = data.draw_integer(low, high)
             result[name] = val
             low_bound = low_bound and val == low
             high_bound = high_bound and val == high
@@ -339,11 +363,11 @@ def timedeltas(
     return TimedeltaStrategy(min_value=min_value, max_value=max_value)
 
 
-@lru_cache(maxsize=None)
+@cache
 def _valid_key_cacheable(tzpath, key):
     assert isinstance(tzpath, tuple)  # zoneinfo changed, better update this function!
     for root in tzpath:
-        if os.path.exists(os.path.join(root, key)):  # pragma: no branch
+        if Path(root).joinpath(key).exists():  # pragma: no branch
             # No branch because most systems only have one TZPATH component.
             return True
     else:  # pragma: no cover
@@ -353,12 +377,7 @@ def _valid_key_cacheable(tzpath, key):
         *package_loc, resource_name = key.split("/")
         package = "tzdata.zoneinfo." + ".".join(package_loc)
         try:
-            try:
-                traversable = resources.files(package) / resource_name
-                return traversable.exists()
-            except (AttributeError, ValueError):
-                # .files() was added in Python 3.9
-                return resources.is_resource(package, resource_name)
+            return (resources.files(package) / resource_name).exists()
         except ModuleNotFoundError:
             return False
 
@@ -388,14 +407,9 @@ def timezone_keys(
 
     .. note::
 
-        The :mod:`python:zoneinfo` module is new in Python 3.9, so you will need
-        to install the :pypi:`backports.zoneinfo` module on earlier versions.
-
-        `On Windows, you will also need to install the tzdata package
+        `The tzdata package is required on Windows
         <https://docs.python.org/3/library/zoneinfo.html#data-sources>`__.
-
-        ``pip install hypothesis[zoneinfo]`` will install these conditional
-        dependencies if and only if they are needed.
+        ``pip install hypothesis[zoneinfo]`` installs it, if and only if needed.
 
     On Windows, you may need to access IANA timezone data via the :pypi:`tzdata`
     package.  For non-IANA timezones, such as Windows-native names or GNU TZ
@@ -406,13 +420,8 @@ def timezone_keys(
     # check_type(bool, allow_alias, "allow_alias")
     # check_type(bool, allow_deprecated, "allow_deprecated")
     check_type(bool, allow_prefix, "allow_prefix")
-    if zoneinfo is None:  # pragma: no cover
-        raise ModuleNotFoundError(
-            "The zoneinfo module is required, but could not be imported.  "
-            "Run `pip install hypothesis[zoneinfo]` and try again."
-        )
 
-    available_timezones = ("UTC",) + tuple(sorted(zoneinfo.available_timezones()))
+    available_timezones = ("UTC", *sorted(zoneinfo.available_timezones()))
 
     # TODO: filter out alias and deprecated names if disallowed
 
@@ -448,21 +457,11 @@ def timezones(*, no_cache: bool = False) -> SearchStrategy["zoneinfo.ZoneInfo"]:
 
     .. note::
 
-        The :mod:`python:zoneinfo` module is new in Python 3.9, so you will need
-        to install the :pypi:`backports.zoneinfo` module on earlier versions.
-
-        `On Windows, you will also need to install the tzdata package
+        `The tzdata package is required on Windows
         <https://docs.python.org/3/library/zoneinfo.html#data-sources>`__.
-
-        ``pip install hypothesis[zoneinfo]`` will install these conditional
-        dependencies if and only if they are needed.
+        ``pip install hypothesis[zoneinfo]`` installs it, if and only if needed.
     """
     check_type(bool, no_cache, "no_cache")
-    if zoneinfo is None:  # pragma: no cover
-        raise ModuleNotFoundError(
-            "The zoneinfo module is required, but could not be imported.  "
-            "Run `pip install hypothesis[zoneinfo]` and try again."
-        )
     return timezone_keys().map(
         zoneinfo.ZoneInfo.no_cache if no_cache else zoneinfo.ZoneInfo
     )
